"""
Utility classes
"""

from datetime import datetime, timedelta
from typing import Any


class CachedItem():
    """ Holds a cached item """

    DEFAULT_LIFETIME_SEC = 10

    def __init__(self, key: Any = None, value: Any = None,
                 timeout_sec=DEFAULT_LIFETIME_SEC):
        self.key = key
        self.value = value
        self._time = datetime.now()
        self._timeout = timedelta(seconds=timeout_sec)

    def expired(self, key: Any = None):
        """ Determines if the token needs to be refreshed """
        return not self.value or (
            self.key == key and
            self._time < datetime.now() - self._timeout
        )



"""
Exposes various high-level operations for the YubiKey.
"""

from base64 import b64encode
from datetime import datetime, timedelta
from getpass import getpass
import json
import sys
from time import time
from typing import List

import requests
from cryptography.hazmat.primitives import serialization
from ykman.descriptor import open_device
from ykman.cli.util import prompt_for_touch
from ykman.piv import \
    ALGO, DEFAULT_MANAGEMENT_KEY, \
    PIN_POLICY, TOUCH_POLICY, SLOT, PivController as YubiKey

from google_yubikey.util import CachedItem

_KEY_ALG = ALGO.RSA2048
_GOOGLE_OAUTH2_TOKEN_ENDPOINT = 'https://oauth2.googleapis.com/token'
DEFAULT_SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
DEFAULT_LIFETIME = 3600


def _log(message: str, file=sys.stderr):
    """ Print information for the user """
    print(message, file=file)


def get_yubikey(stream=sys.stderr):
    """ Sets up YubiKey communication """
    _log('Connecting to the YubiKey...', stream)
    dev = open_device()
    yubikey = YubiKey(dev.driver)
    _log('Connected', stream)
    return yubikey


_CACHED_PIN = CachedItem()
_CACHED_MGMT_KEY = CachedItem()


def authenticate(yubikey: YubiKey, prompt_management_key: bool,
                 cache_lifetime=CachedItem.DEFAULT_LIFETIME_SEC, stream=sys.stderr):
    """ Authenticates user to the YubiKey """
    global _CACHED_PIN, _CACHED_MGMT_KEY  # pylint: disable=global-statement

    _log('Authenticating to the YubiKey...', stream)

    pin = _CACHED_PIN.value
    if _CACHED_PIN.expired():
        pin = getpass('Enter PIN: ', stream)
        _CACHED_PIN = CachedItem(None, pin, cache_lifetime)
    yubikey.verify(pin, touch_callback=prompt_for_touch)

    mgmt_key = _CACHED_MGMT_KEY.value
    if not prompt_management_key:
        mgmt_key = DEFAULT_MANAGEMENT_KEY
    elif _CACHED_MGMT_KEY.expired():
        mgmt_key = getpass('Enter management key: ', stream)
        _CACHED_MGMT_KEY = CachedItem(None, mgmt_key, cache_lifetime)
    yubikey.authenticate(mgmt_key, touch_callback=prompt_for_touch)

    _log('Authenticated', stream)


def gen_private_key(yubikey: YubiKey, slot: SLOT, prompt_management_key: bool,
                    pin_policy: PIN_POLICY, touch_policy: TOUCH_POLICY,
                    subject: str, valid_days: int):
    """ Generates a private key and certificate on the YubiKey """
    authenticate(yubikey, prompt_management_key)

    _log('Generating private key on YubiKey...')
    public_key = yubikey.generate_key(
        slot.value, _KEY_ALG, pin_policy.value, touch_policy.value,
    )

    _log('Generating certificate on YubiKey...')
    start = datetime.now()
    end = start + timedelta(days=valid_days)
    yubikey.generate_self_signed_certificate(
        slot.value, public_key, subject, start, end,
        touch_callback=prompt_for_touch,
    )
    return get_public_key(yubikey, slot)


def get_public_key(yubikey: YubiKey, slot: SLOT):
    """ Reads public key from YubiKey """
    cert = yubikey.read_certificate(slot.value)
    return cert.public_bytes(serialization.Encoding.PEM)


def _b64encode_str(bbytes: bytes):
    """ Encodes bytes as base64 string """
    return b64encode(bbytes).decode('utf-8')


def _json_b64encode(obj: dict):
    """ Converts a dict to a base64-encoded JSON string """
    json_str = json.dumps(obj, separators=(',', ':')).encode('utf-8')
    return _b64encode_str(json_str)


def _get_jwt(yubikey: YubiKey, slot: int, prompt_management_key: bool,
             service_account_email: str, audience: str, scopes: List[str],
             token_lifetime: int, cache_lifetime: int, stream=sys.stderr):
    """ Generates a general-purpose Google JWT with a YubiKey """
    authenticate(yubikey, prompt_management_key, cache_lifetime, stream)

    iat = time()
    header = {
        'typ': 'JWT',
        'alg': 'RS256',
    }
    payload = {
        'iss': service_account_email,
        'aud': audience,
        'iat': iat,
        'exp': iat + token_lifetime,
    }
    if scopes:
        payload['scope'] = ' '.join(scopes)

    msg = f'{_json_b64encode(header)}.{_json_b64encode(payload)}'
    sig = yubikey.sign(slot, _KEY_ALG, msg.encode('utf-8'))
    sig = _b64encode_str(sig)

    return f'{msg}.{sig}'


def get_id_token(yubikey: YubiKey, slot: int, prompt_management_key: bool,
                 service_account_email: str, audience: str, token_lifetime: int,
                 cache_lifetime=CachedItem.DEFAULT_LIFETIME_SEC, stream=sys.stderr):
    """ Generates a Google ID token with a YubiKey """
    if not audience:
        raise ValueError('ID tokens must use a non-empty audience')
    return _get_jwt(
        yubikey, slot, prompt_management_key,
        service_account_email, audience, [],
        token_lifetime, cache_lifetime, stream,
    )


def get_access_token(yubikey: YubiKey, slot: int, prompt_management_key: bool,
                     service_account_email: str, scopes: List[str], token_lifetime: int,
                     cache_lifetime=CachedItem.DEFAULT_LIFETIME_SEC, stream=sys.stderr):
    """ Generates a Google Access token with a YubiKey """
    assertion = _get_jwt(
        yubikey, slot, prompt_management_key,
        service_account_email, _GOOGLE_OAUTH2_TOKEN_ENDPOINT, scopes,
        token_lifetime, cache_lifetime, stream,
    )
    response = requests.post(
        url=_GOOGLE_OAUTH2_TOKEN_ENDPOINT,
        data={
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'assertion': assertion,
        },
    )
    if not response.ok:
        raise RuntimeError(response.json()['error_description'])

    return response.json()



#!/usr/bin/env python3
"""
Command-line interface to set up and use a YubiKey
as a Google Service Account key.
"""

import argparse
from enum import Enum
import logging

from ykman.piv import PIN_POLICY, TOUCH_POLICY, SLOT

from google_yubikey.device import \
    get_yubikey, gen_private_key, get_access_token, get_id_token,\
    DEFAULT_LIFETIME, DEFAULT_SCOPES
from google_yubikey.metadata import get_gce_metadata
from google_yubikey.util import CachedItem


class ArgEnum(Enum):
    """ Enum for command-line argument choices """

    def __str__(self):
        return str(self.name).lower().replace('_', '-')

    @staticmethod
    def from_str(clazz):
        """ Converts value from string """
        def convert(str_value: str):
            try:
                return clazz[str_value.upper().replace('-', '_')]
            except KeyError as error:
                raise ValueError() from error
        return convert


class Action(ArgEnum):
    """ Action type """
    GENERATE_KEY = 1
    TOKEN = 2
    SERVE = 3


class Slot(ArgEnum):
    """ YubiKey slot type """
    AUTHENTICATION = SLOT.AUTHENTICATION.value
    CARD_MANAGEMENT = SLOT.CARD_MANAGEMENT.value
    SIGNATURE = SLOT.SIGNATURE.value
    KEY_MANAGEMENT = SLOT.KEY_MANAGEMENT.value
    CARD_AUTH = SLOT.CARD_AUTH.value
    ATTESTATION = SLOT.ATTESTATION.value


class PinPolicy(ArgEnum):
    """ YubiKey pin policy """
    DEFAULT = PIN_POLICY.DEFAULT.value
    NEVER = PIN_POLICY.NEVER.value
    ONCE = PIN_POLICY.ONCE.value
    ALWAYS = PIN_POLICY.ALWAYS.value


class TouchPolicy(ArgEnum):
    """ YubiKey touch policy """
    DEFAULT = TOUCH_POLICY.DEFAULT.value
    NEVER = TOUCH_POLICY.NEVER.value
    ALWAYS = TOUCH_POLICY.ALWAYS.value
    CACHED = TOUCH_POLICY.CACHED.value


class TokenType(ArgEnum):
    """ Google token type """
    ID = 1
    ACCESS = 2


class Verbosity(ArgEnum):
    """ Google token type """
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NONE = logging.NOTSET


def parse_args():
    """ Parses command-line args """

    # top-level args
    parser = argparse.ArgumentParser(
        description='Generate a Google Service Account token with YubiKey',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '-k', '--key-slot', dest='slot', type=Slot.from_str(Slot),
        help='YubiKey slot', default=Slot.AUTHENTICATION, choices=list(Slot),
    )
    subparsers = parser.add_subparsers(
        dest='action', required=True,
        help='Action',
    )

    # "generate-key" action
    parser_generate_key = subparsers.add_parser(
        str(Action.GENERATE_KEY),
        help='Generate private key on the YubiKey '
        '(can be used for many Service Accounts)',
    )
    parser_generate_key.add_argument(
        '-s', '--subject',
        help='Subject common name (CN) for the key', default='yubikey',
    )
    parser_generate_key.add_argument(
        '-d', '--valid-days', type=int,
        help='Number of days until the key expires', default=365,
    )
    parser_generate_key.add_argument(
        '-p', '--pin-policy', type=PinPolicy.from_str(PinPolicy),
        help='YubiKey PIN policy', default=PinPolicy.DEFAULT, choices=list(PinPolicy),
    )
    parser_generate_key.add_argument(
        '-t', '--touch-policy', type=TouchPolicy.from_str(TouchPolicy),
        help='YubiKey touch policy', default=TouchPolicy.DEFAULT, choices=list(TouchPolicy),
    )
    parser_generate_key.add_argument(
        '-m', '--prompt-management-key', action='store_true',
        help='Prompt for management key',
    )

    # "token" action
    parser_token = subparsers.add_parser(
        str(Action.TOKEN),
        help='Generate a token',
    )
    parser_token.add_argument(
        '-a', '--service-account-email', required=True,
        help='Service Account email',
    )
    parser_token.add_argument(
        '-t', '--token-type', type=TokenType.from_str(TokenType), choices=list(TokenType),
        help='Token type', default=TokenType.ACCESS,
    )
    parser_token.add_argument(
        '-d', '--audience',
        help='Audience for ID token',
    )
    parser_token.add_argument(
        '-s', '--scopes', nargs='*', default=DEFAULT_SCOPES,
        help='Google Cloud access token scope(s)',
    )
    parser_token.add_argument(
        '-l', '--token-lifetime', type=int, default=DEFAULT_LIFETIME,
        help='Token lifetime, in seconds',
    )
    parser_token.add_argument(
        '-m', '--prompt-management-key', action='store_true',
        help='Prompt for management key',
    )

    # "serve" action
    parser_serve = subparsers.add_parser(
        str(Action.SERVE),
        help='Start a server that provides application default credentials',
    )
    parser_serve.add_argument(
        '-n', '--numeric-project-id', type=int, required=True,
        help='Google Cloud numeric project id',
    )
    parser_serve.add_argument(
        '-a', '--service-account-email', required=True,
        help='Service Account email',
    )
    parser_serve.add_argument(
        '-l', '--token-lifetime', type=int, default=DEFAULT_LIFETIME,
        help='Token lifetime, in seconds',
    )
    parser_serve.add_argument(
        '-c', '--cache-lifetime', type=int, default=CachedItem.DEFAULT_LIFETIME_SEC,
        help='Token/PIN cache lifetime, in seconds',
    )
    parser_serve.add_argument(
        '-m', '--prompt-management-key', action='store_true',
        help='Prompt for management key',
    )
    parser_serve.add_argument(
        '-v', '--verbosity', type=Verbosity.from_str(Verbosity), choices=list(Verbosity),
        help='Log verbosity level', default=Verbosity.INFO,
    )

    return parser.parse_args()


def main():
    """ Main entrypoint """
    args = parse_args()

    if args.action == str(Action.GENERATE_KEY):
        public_key = gen_private_key(
            get_yubikey(), args.slot, args.prompt_management_key,
            args.pin_policy, args.touch_policy,
            args.subject, args.valid_days,
        )
        print(public_key.decode('utf-8'))
    elif args.action == str(Action.TOKEN):
        if args.token_type == TokenType.ACCESS:
            print(get_access_token(
                get_yubikey(), args.slot.value, args.prompt_management_key,
                args.service_account_email, args.scopes, args.token_lifetime,
            )['access_token'])
        else:
            print(get_id_token(
                get_yubikey(), args.slot.value, args.prompt_management_key,
                args.service_account_email, args.audience, args.token_lifetime,
            ))
    elif args.action == str(Action.SERVE):
        get_gce_metadata(
            args.slot, args.prompt_management_key, args.numeric_project_id,
            args.service_account_email, args.token_lifetime, args.cache_lifetime,
            args.verbosity.name,
        ).run()


if __name__ == "__main__":
    main()



"""
Emulates a GCE instance metadata server
to seamlessly generate Service Account credentials
for existing applications with a YubiKey.

It must be called with superuser privileges,
to be allowed to set up the link-local IP alias
for the instance metadata server, and
to serve its endpoint on port 80 (the privileges
are dropped for the actual request handlers, however).
"""

from abc import ABCMeta, abstractmethod
import atexit
import logging
import platform
from subprocess import PIPE, run
import sys
from typing import Type

from flask import Flask, Response, abort, redirect, request, url_for
from python_hosts import Hosts, HostsEntry
from werkzeug.exceptions import HTTPException
from ykman.piv import SLOT

from google_yubikey.device import \
    get_access_token, get_id_token, get_yubikey, \
    DEFAULT_SCOPES
from google_yubikey.util import CachedItem


class GCEMetadata:
    """ Interface for setting up GCE Metadata server in OS-specific way """
    __metaclass__ = ABCMeta

    HOSTS = ['metadata', 'metadata.google.internal']
    IP = '169.254.169.254'

    def __init__(self, slot: SLOT, prompt_management_key: bool,
                 numeric_project_id: int, service_account_email: str,
                 token_lifetime: int, cache_lifetime: int, verbosity: str):
        self.slot = slot
        self.prompt_management_key = prompt_management_key
        self.numeric_project_id = numeric_project_id
        self.service_account_email = service_account_email
        self.token_lifetime = token_lifetime
        self.cache_lifetime = cache_lifetime
        self.verbosity = verbosity

        self.add_ip()
        atexit.register(self.delete_ip)

        self.add_host()
        atexit.register(self.delete_host)

    @abstractmethod
    def add_ip(self):
        """ Add the server IP alias """
        raise NotImplementedError

    @abstractmethod
    def delete_ip(self):
        """ Delete the server IP alias """
        raise NotImplementedError

    @staticmethod
    def add_host():
        """ Add hosts entry for the server IP """
        hosts = Hosts()
        hosts.add([
            HostsEntry(
                entry_type='ipv4',
                address=GCEMetadata.IP,
                names=GCEMetadata.HOSTS,
            ),
        ])
        hosts.write()

    @staticmethod
    def delete_host():
        """ Delete hosts entry for the server IP """
        hosts = Hosts()
        hosts.remove_all_matching(GCEMetadata.IP)
        hosts.write()

    def run(self):
        """ Runs the server """
        is_debug = getattr(logging, self.verbosity) <= logging.DEBUG
        stderr = sys.stderr if is_debug else PIPE
        server = run([
            'uwsgi',
            '--http', '=0',
            '--shared-socket', GCEMetadata.IP + ':80',
            '--uid', 'nobody', '--gid', 'nobody',
            '--wsgi', 'google_yubikey.metadata:create_uwsgi_app()',
            '--set', f'slot={self.slot.value}',
            '--set', f'prompt_management_key={self.prompt_management_key}',
            '--set', f'numeric_project_id={self.numeric_project_id}',
            '--set', f'service_account_email={self.service_account_email}',
            '--set', f'token_lifetime={self.token_lifetime}',
            '--set', f'cache_lifetime={self.cache_lifetime}',
            '--set', f'verbosity={self.verbosity}',
            '--processes', '1',
            '--honour-stdin',
        ], stdout=sys.stdout, stderr=stderr, text=True, check=False)

        if not is_debug and server.returncode != 0:
            print(server.stderr)


class GCEMetadataLinux(GCEMetadata):
    """ Implements GCEMetadata for Linux """

    _INTERFACE = 'lo:0'

    def add_ip(self):
        run(
            ['ifconfig', self._INTERFACE, GCEMetadata.IP],
            stderr=sys.stderr, check=True,
        )

    def delete_ip(self):
        run(
            ['ifconfig', self._INTERFACE, 'down'],
            stderr=sys.stderr, check=True,
        )


class GCEMetadataMacOS(GCEMetadata):
    """ Implements GCEMetadata for macOS """

    _INTERFACE = 'lo0'

    def add_ip(self):
        run(
            ['ifconfig', self._INTERFACE, 'alias', GCEMetadata.IP],
            stderr=sys.stderr, check=True,
        )

    def delete_ip(self):
        run(
            ['ifconfig', self._INTERFACE, '-alias', GCEMetadata.IP],
            stderr=sys.stderr, check=True,
        )


def get_gce_metadata(slot: SLOT, prompt_management_key: bool,
                     numeric_project_id: int, service_account_email: str,
                     token_lifetime: int, cache_lifetime: int,
                     verbosity: str) -> GCEMetadata:
    """ Returns GCEMetadata instance for your OS """
    os_name = platform.system()
    metadata_type: Type[GCEMetadata]
    if os_name == 'Linux':
        metadata_type = GCEMetadataLinux
    elif os_name == 'Darwin':
        metadata_type = GCEMetadataMacOS
    else:
        raise NotImplementedError('Sorry, your OS is not supported yet')
    return metadata_type(
        slot, prompt_management_key, numeric_project_id,
        service_account_email, token_lifetime, cache_lifetime, verbosity,
    )


class UWSGIOpts:
    """ Parses options passed through uWSGI """

    def __init__(self):
        from uwsgi import opt  # pylint: disable=import-error,import-outside-toplevel

        self._opt = opt
        self.slot = int(self._get('slot'))
        self.prompt_management_key = \
            self._get('prompt_management_key') == 'True'
        self.service_account_email = self._get('service_account_email')
        self.numeric_project_id = self._get('numeric_project_id')
        self.project_id = self.service_account_email \
            .split('@')[1].split('.')[0]
        self.token_lifetime = int(self._get('token_lifetime'))
        self.cache_lifetime = int(self._get('cache_lifetime'))
        self.verbosity = self._get('verbosity')

    def _get(self, name: str):
        return self._opt[name].decode('utf-8')


def _get_log(name: str, level: str):
    """ Sets up a logger """
    log = logging.getLogger(name)
    log_handler = logging.StreamHandler(sys.stdout)
    log_handler.setFormatter(logging.Formatter(
        '[%(levelname)s][%(asctime)s][%(name)s]: %(message)s',
    ))
    log.addHandler(log_handler)
    log.setLevel(level)
    return log


_CACHED_TOKEN = CachedItem()
_CACHED_IDENTITY = CachedItem()

_PROJECT_ROOT = '/computeMetadata/v1/project'
_SA_ROOT = '/computeMetadata/v1/instance/service-accounts'


def create_uwsgi_app():
    """ Sets up uWSGI app for GCE metadata server """
    opts = UWSGIOpts()
    log = _get_log('gce_metadata', opts.verbosity)
    yubikey = get_yubikey(sys.stdout)
    app = Flask(__name__)

    @app.before_request
    def _check_request_headers():
        log.debug(request.method + ' ' + request.path)
        if request.path == '/':
            return None
        headers = request.headers
        if headers.get('X-Forwarded-For') or \
                headers.get('Metadata-Flavor') != 'Google' or \
                headers.get('Host') not in GCEMetadata.HOSTS + [GCEMetadata.IP]:
            return abort(401)
        return None

    @app.after_request
    def _add_response_headers(response: Response):
        headers = response.headers
        headers.set('Server', 'Metadata Server for VM')
        headers.set('Metadata-Flavor', 'Google')
        headers.set('X-XSS-Protection', '1')
        headers.set('X-Frame-Options', 'SAMEORIGIN')
        if response.mimetype == 'text/html':
            headers.set('Content-Type', 'application/text')
        return response

    @app.errorhandler(Exception)
    def _log_error(error: Exception):
        log.error(error)
        if isinstance(error, HTTPException):
            return error.description, error.code
        msg = error
        if len(error.args) > 0:
            msg = error.args[0]
        if msg == 'Incorrect PIN':
            return msg, 401
        return msg, 500

    @app.route('/')
    def _get_root():
        return 'computeMetadata/'

    @app.route(_PROJECT_ROOT + '/project-id')
    def _get_project_id():
        return opts.project_id

    @app.route(_PROJECT_ROOT + '/numeric-project-id')
    def _get_numeric_project_id():
        return opts.numeric_project_id

    @app.route(_SA_ROOT)
    def _get_service_accounts():
        return redirect(url_for('_get_service_accounts_root'), 301)

    @app.route(_SA_ROOT + '/')
    def _get_service_accounts_root():
        return '/\n'.join([
            'default',
            opts.service_account_email,
            '',
        ])

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/')
    @app.route(_SA_ROOT + '/default/')
    def _get_sa_index():
        return {
            'aliases': 'default',
            'email': opts.service_account_email,
            'scopes': DEFAULT_SCOPES,
        }

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/aliases')
    @app.route(_SA_ROOT + '/default/aliases')
    def _get_aliases():
        return 'default'

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/email')
    @app.route(_SA_ROOT + '/default/email')
    def _get_email():
        return opts.service_account_email

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/identity')
    @app.route(_SA_ROOT + '/default/identity')
    def _get_id_token():
        global _CACHED_IDENTITY  # pylint: disable=global-statement

        audience = request.args.get('audience')
        if not audience:
            abort(400, 'non-empty audience parameter required')
        if not _CACHED_IDENTITY.expired(audience):
            return _CACHED_IDENTITY.value
        response = get_id_token(
            yubikey, opts.slot, opts.prompt_management_key,
            opts.service_account_email, audience,
            opts.token_lifetime, opts.cache_lifetime, sys.stdout,
        )
        _CACHED_IDENTITY = CachedItem(audience, response, opts.cache_lifetime)
        return response

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/scopes')
    @app.route(_SA_ROOT + '/default/scopes')
    def _get_scopes():
        return '\n'.join(DEFAULT_SCOPES + [''])

    @app.route(_SA_ROOT + f'/{opts.service_account_email}/token')
    @app.route(_SA_ROOT + '/default/token')
    def _get_access_token():
        global _CACHED_TOKEN  # pylint: disable=global-statement

        scopes = request.args.get('scopes')
        if scopes:
            scopes = scopes.split(',')
        else:
            scopes = DEFAULT_SCOPES
        if not _CACHED_TOKEN.expired(scopes):
            return _CACHED_TOKEN.value

        response = get_access_token(
            yubikey, opts.slot, opts.prompt_management_key,
            opts.service_account_email, scopes,
            opts.token_lifetime, opts.cache_lifetime, sys.stdout,
        )
        _CACHED_TOKEN = CachedItem(scopes, response, opts.cache_lifetime)
        return response

    return app
